"""Checks for the django-help package."""

from django.conf import settings
from django.core.checks import Info
from django.core.checks import Warning
from django.core.checks import register
from django.db import models

from django_help import app_settings


@register()
def check_extra_languages_in_settings(app_configs, **kwargs):  # pylint: disable=W0613  # pylint: disable=W0613
    """Check that languages in EXTRA_LANGUAGES are available in settings.LANGUAGES."""

    errors = []
    available_languages = dict(settings.LANGUAGES)

    for lang_code in app_settings.EXTRA_LANGUAGES:
        if lang_code not in available_languages:
            errors.append(
                Warning(
                    f"Language '{lang_code}' specified in EXTRA_LANGUAGES is not available in " "settings.LANGUAGES.",
                    hint="Add the language to settings.LANGUAGES or remove it from EXTRA_LANGUAGES.",
                    id="django_help.W001",
                )
            )

    return errors


@register()
def check_languages_in_extra_languages(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that languages in settings.LANGUAGES are specified in EXTRA_LANGUAGES except for LANGUAGE_CODE."""

    messages = []
    available_languages = dict(settings.LANGUAGES)

    for lang_code, lang_name in available_languages.items():
        if lang_code not in app_settings.EXTRA_LANGUAGES and lang_code != settings.LANGUAGE_CODE:
            messages.append(
                Info(
                    f"Language '{lang_code}' ({lang_name}) from settings.LANGUAGES is not specified in "
                    "EXTRA_LANGUAGES.",
                    hint="Consider adding the language to EXTRA_LANGUAGES if it should be translatable.",
                    id="django_help.I001",
                )
            )

    return messages


@register()
def check_authorize_user_to_view_article(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that AUTHORIZE_USER_TO_VIEW_ARTICLE is a callable."""
    errors = []
    authorize_user_to_view_article = app_settings.AUTHORIZE_USER_TO_VIEW_ARTICLE

    if not callable(authorize_user_to_view_article):
        errors.append(
            Warning(
                "AUTHORIZE_USER_TO_VIEW_ARTICLE is not a callable.",
                hint="AUTHORIZE_USER_TO_VIEW_ARTICLE should be a callable function that takes a request and "
                "an article as arguments.",
                id="django_help.W002",
            )
        )

    return errors


@register()
def check_intended_entity_type(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that INTENDED_ENTITY_TYPE is a dictionary."""
    errors = []
    intended_entity_type = app_settings.INTENDED_ENTITY_TYPE

    if not isinstance(intended_entity_type, dict):
        errors.append(
            Warning(
                "INTENDED_ENTITY_TYPE is not a dictionary.",
                hint="INTENDED_ENTITY_TYPE should be a dictionary of intended entity types.",
                id="django_help.W003",
            )
        )

    return errors


@register()
def check_base_manager(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that BASE_MANAGER is a subclass of django.db.models.Manager."""
    errors = []
    base_manager = app_settings.BASE_MANAGER

    if not issubclass(base_manager, models.Manager):
        errors.append(
            Warning(
                "BASE_MANAGER is not a subclass of django.db.models.Manager.",
                hint="BASE_MANAGER should be a subclass of django.db.models.Manager.",
                id="django_help.W004",
            )
        )

    return errors


@register()
def check_base_queryset(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that BASE_QUERYSET is a subclass of django.db.models.QuerySet."""
    errors = []
    base_queryset = app_settings.BASE_QUERYSET

    if not issubclass(base_queryset, models.QuerySet):
        errors.append(
            Warning(
                "BASE_QUERYSET is not a subclass of django.db.models.QuerySet.",
                hint="BASE_QUERYSET should be a subclass of django.db.models.QuerySet.",
                id="django_help.W005",
            )
        )

    return errors


@register()
def check_base_model(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that BASE_MODEL is a subclass of django.db.models.Model."""
    errors = []
    base_model = app_settings.BASE_MODEL

    if not issubclass(base_model, models.Model):
        errors.append(
            Warning(
                "BASE_MODEL is not a subclass of django.db.models.Model.",
                hint="BASE_MODEL should be a subclass of django.db.models.Model.",
                id="django_help.W006",
            )
        )

    return errors


@register()
def check_base_foreign_key(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that BASE_FOREIGN_KEY is a subclass of django.db.models.ForeignKey."""
    errors = []
    base_foreign_key = app_settings.BASE_FOREIGN_KEY_FIELD

    if not issubclass(base_foreign_key, models.ForeignKey):
        errors.append(
            Warning(
                "BASE_FOREIGN_KEY is not a subclass of django.db.models.ForeignKey.",
                hint="BASE_FOREIGN_KEY should be a subclass of django.db.models.ForeignKey.",
                id="django_help.W007",
            )
        )

    return errors


@register()
def check_extra_languages(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that EXTRA_LANGUAGES is a dictionary."""
    errors = []
    extra_languages = app_settings.EXTRA_LANGUAGES

    if not isinstance(extra_languages, dict):
        errors.append(
            Warning(
                "EXTRA_LANGUAGES is not a dictionary.",
                hint="EXTRA_LANGUAGES should be a dictionary of extra languages.",
                id="django_help.W008",
            )
        )

    return errors


@register()
def check_extra_languages_values(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that values in EXTRA_LANGUAGES are dictionaries."""
    errors = []
    extra_languages = app_settings.EXTRA_LANGUAGES

    for lang_code, lang_settings in extra_languages.items():
        if not isinstance(lang_settings, dict):
            errors.append(
                Warning(
                    f"Value for language '{lang_code}' in EXTRA_LANGUAGES is not a dictionary.",
                    hint="Values in EXTRA_LANGUAGES should be dictionaries.",
                    id="django_help.W009",
                )
            )

    return errors


@register()
def check_extra_languages_values_keys(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that keys in EXTRA_LANGUAGES values are valid."""
    errors = []
    extra_languages = app_settings.EXTRA_LANGUAGES

    for lang_code, lang_settings in extra_languages.items():
        for key in lang_settings:
            if key not in ("blank",):
                errors.append(
                    Warning(
                        f"Invalid key '{key}' in EXTRA_LANGUAGES value for language '{lang_code}'.",
                        hint="Valid keys are 'blank'.",
                        id="django_help.W010",
                    )
                )

    return errors


@register()
def check_extra_languages_values_blank(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that 'blank' key in EXTRA_LANGUAGES values is a boolean."""
    errors = []
    extra_languages = app_settings.EXTRA_LANGUAGES

    for lang_code, lang_settings in extra_languages.items():
        blank = lang_settings.get("blank")

        if not isinstance(blank, bool):
            errors.append(
                Warning(
                    f"Value for 'blank' key in EXTRA_LANGUAGES value for language '{lang_code}' is not a " "boolean.",
                    hint="'blank' should be a boolean.",
                    id="django_help.W011",
                )
            )

    return errors


@register()
def check_intended_entity_type_values(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that values in INTENDED_ENTITY_TYPE are tuples."""
    errors = []
    intended_entity_type = app_settings.INTENDED_ENTITY_TYPE

    for entity_type, entity_settings in intended_entity_type.items():
        if not isinstance(entity_settings, tuple):
            errors.append(
                Warning(
                    f"Value for entity type '{entity_type}' in INTENDED_ENTITY_TYPE is not a tuple.",
                    hint="Values in INTENDED_ENTITY_TYPE should be tuples.",
                    id="django_help.W012",
                )
            )

    return errors


@register()
def check_intended_entity_type_values_length(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that values in INTENDED_ENTITY_TYPE tuples have a length of 2."""
    errors = []
    intended_entity_type = app_settings.INTENDED_ENTITY_TYPE

    for entity_type, entity_settings in intended_entity_type.items():
        if len(entity_settings) != 2:
            errors.append(
                Warning(
                    f"Value for entity type '{entity_type}' in INTENDED_ENTITY_TYPE does not have a length " "of 2.",
                    hint="Values in INTENDED_ENTITY_TYPE should have a length of 2.",
                    id="django_help.W013",
                )
            )

    return errors


@register()
def check_intended_entity_type_values_types(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that values in INTENDED_ENTITY_TYPE tuples are strings."""
    errors = []
    intended_entity_type = app_settings.INTENDED_ENTITY_TYPE

    for entity_type, entity_settings in intended_entity_type.items():
        if not all(isinstance(value, str) for value in entity_settings):
            errors.append(
                Warning(
                    f"Values for entity type '{entity_type}' in INTENDED_ENTITY_TYPE are not strings.",
                    hint="Values in INTENDED_ENTITY_TYPE should be strings.",
                    id="django_help.W014",
                )
            )

    return errors


@register()
def check_intended_entity_type_values_unique(app_configs, **kwargs):  # pylint: disable=W0613
    """Check that values in INTENDED_ENTITY_TYPE tuples are unique."""
    errors = []
    intended_entity_type = app_settings.INTENDED_ENTITY_TYPE
    values = [value for _, value in intended_entity_type.values()]

    if len(values) != len(set(values)):
        errors.append(
            Warning(
                "Values in INTENDED_ENTITY_TYPE tuples are not unique.",
                hint="Values in INTENDED_ENTITY_TYPE should be unique.",
                id="django_help.W015",
            )
        )

    return errors



"""Validation utilities for the Django Help app."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Optional

from django.http import Http404
from django.http import HttpRequest


if TYPE_CHECKING:
    from django_help.models import DjangoHelpArticle


logger = logging.getLogger(__name__)


def user_authorized_to_view(request: HttpRequest, article: DjangoHelpArticle) -> Optional[Http404]:
    """Raises a 404 error if the user is not allowed to view the article."""
    if not request.user.is_staff:
        return

    if not (article.category and article.category.public and article.public):
        logger.warning("User %s attempted to view non-public article %s", request.user, article)
        raise Http404



"""Models for django_help."""

import logging
from typing import List
from typing import Union

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.http import HttpRequest
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _
from markdownx.models import MarkdownxField
from taggit.managers import TaggableManager
from taggit.models import ItemBase
from taggit.models import TagBase
from translated_fields import TranslatedField

from django_help.app_settings import BASE_FOREIGN_KEY_FIELD
from django_help.app_settings import BASE_MANAGER
from django_help.app_settings import BASE_MODEL
from django_help.app_settings import BASE_QUERYSET
from django_help.app_settings import EXTRA_LANGUAGES
from django_help.choices import IntendedEntityType
from django_help.utils.regex import get_path_regex


logger = logging.getLogger(__name__)

language_codes = [code for code, _ in settings.LANGUAGES]


def get_only_filters():
    """Return the fields to select in a queryset."""
    current_language_code = get_language()
    return [
        f"title_{current_language_code}",
        f"subtitle_{current_language_code}",
        "slug",
        "modified",
    ]


class Tag(TagBase, BASE_MODEL):
    """Customized Taggit Tag model."""

    class Meta(BASE_MODEL.Meta if hasattr(BASE_MODEL, "Meta") else object):
        """Meta options for Tag."""

        verbose_name = _("tag")
        verbose_name_plural = _("tags")


class DjangoHelpCategoryQuerySet(BASE_QUERYSET):
    """QuerySet for DjangoHelpCategory."""

    def public(self):
        """Return only categories which are public."""
        return self.filter(public=True)

    def private(self):
        """Return only categories which are not public."""
        return self.filter(public=False)

    def intended_for_any(self):
        """Return categories intended for any entity."""
        return self.filter(intended_entity_type=IntendedEntityType.ANY)


class DjangoHelpCategoryManager(BASE_MANAGER):
    """Manager for DjangoHelpCategory."""

    def get_queryset(self):
        """Return the queryset for this manager."""
        return super().get_queryset().all()


class DjangoHelpCategory(BASE_MODEL):
    """Stores categories."""

    title = TranslatedField(
        models.CharField(
            _("Title"),
            max_length=30,  # Specifically to fit the cards in the django_help index page
            help_text=_("The title of this category."),
            blank=True,
        ),
        EXTRA_LANGUAGES,
    )
    subtitle = TranslatedField(
        models.CharField(
            _("Subtitle"),
            max_length=70,
            blank=True,
            help_text=_("A subtitle for this category."),
        ),
        EXTRA_LANGUAGES,
    )
    slug = models.SlugField(
        _("Slug"),
        max_length=50,
        unique=True,
        help_text=_("A web address friendly version of the title."),
    )
    description = TranslatedField(
        models.CharField(
            _("Description"),
            max_length=200,
            blank=True,
            help_text=_("A description of this category."),
        ),
        EXTRA_LANGUAGES,
    )
    icon = models.CharField(
        max_length=50,
        blank=True,
        help_text=_("The icon and text color to use for this category."),
        default="fa-circle-info text-success",
    )
    public = models.BooleanField(
        default=True,
        help_text=_("Check this box to make this category public."),
    )

    intended_entity_type = models.CharField(
        max_length=20,
        choices=IntendedEntityType.choices,
        default=IntendedEntityType.ANY,
        help_text=_("The type of entity this category applies to."),
    )

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    objects = DjangoHelpCategoryManager.from_queryset(DjangoHelpCategoryQuerySet)()

    class Meta(BASE_MODEL.Meta if hasattr(BASE_MODEL, "Meta") else object):
        """Meta options for DjangoHelpCategory."""

        verbose_name = _("DjangoHelp Category")
        verbose_name_plural = _("DjangoHelp Categories")
        ordering = [f"title_{settings.LANGUAGE_CODE}"]  # Default to the primary language
        indexes = [
            models.Index(fields=["slug"]),
        ]

    def __str__(self):
        return self.title

    @property
    def short_description(self, length=50):
        """Return a short description of this category."""
        return self.description[:length] + "..." if len(self.description) > length else self.description

    def add_article(self, article):
        """Add the specified article to this category."""
        article.category = self
        article.save()

    @staticmethod
    def remove_article(article):
        """Remove the specified article from its category."""
        article.category = None
        article.save()

    @property
    def article_count(self):
        """Return the number of article in this category."""
        return self.articles.count()


class TaggedArticles(ItemBase, BASE_MODEL):
    """Stores tagged articles."""

    content_object = BASE_FOREIGN_KEY_FIELD(
        "django_help.DjangoHelpArticle",
        on_delete=models.CASCADE,
    )
    tag = BASE_FOREIGN_KEY_FIELD(
        Tag,
        related_name="%(app_label)s_%(class)s_items",
        on_delete=models.CASCADE,
    )

    class Meta(BASE_MODEL.Meta if hasattr(BASE_MODEL, "Meta") else object):
        """Meta options for TaggedArticles."""

        verbose_name = _("Tagged Article")
        verbose_name_plural = _("Tagged Articles")


class DjangoHelpArticleQuerySet(BASE_QUERYSET):
    """QuerySet for DjangoHelpArticle."""

    def public(self):
        """Return only article which is public."""
        return self.filter(public=True)

    def private(self):
        """Return only article which is not public."""
        return self.filter(public=False)

    def intended_for_any(self):
        """Return article intended for any entity."""
        return self.filter(intended_entity_type=IntendedEntityType.ANY)

    def search(self, search_terms: List[str]):
        """Search for article matching the specified query in the current language."""
        if len(search_terms) == 0:
            return self.none()

        # For each search term, create filters in the current language.
        title_filters = {f"title_{get_language()}__icontains": search_term for search_term in search_terms}
        subtitle_filters = {f"subtitle_{get_language()}__icontains": search_term for search_term in search_terms}
        article_content_filters = {
            f"article_content_{get_language()}__icontains": search_term for search_term in search_terms
        }
        tags_filters = {"tags__name__icontains": search_term for search_term in search_terms}

        filters = Q(**title_filters) | Q(**subtitle_filters) | Q(**article_content_filters) | Q(**tags_filters)
        logger.debug("Called search in DjangoHelpArticleQuerySet with filters: %s", filters)
        return self.filter(filters)

    def get_for_tag(self, tag: str, multiple=False):
        """Return article(s) matching the specified tag.

        If `single` is True (default), return only the first article.
        """
        qs = self.filter(tags__name__in=[tag])
        if multiple:
            return qs
        return qs.first()

    def get_for_category(self, category: str, multiple=False):
        """Return article(s) matching the specified category.

        If `single` is True (default), return only the first article.
        """
        title_filters = {f"category__title_{code}__icontains": category.lower() for code in language_codes}
        slug_filters = {"category__slug__icontains": category.lower()}

        filters = Q(**title_filters) | Q(**slug_filters)
        logger.debug("Called get_for_category in DjangoHelpArticleQuerySet with filters: %s", filters)
        qs = self.filter(filters)
        if multiple:
            return qs
        return qs.first()

    def get_for_slug(self, slug: str, multiple=False):
        """Return article(s) matching the specified slug.

        If `single` is True (default), return only the first article.
        """
        qs = self.filter(slug__iexact=slug.lower())
        if multiple:
            return qs
        return qs.first()

    def get_for_path(self, path: Union[str, HttpRequest], multiple=False):
        """Return article(s) matching the specified path.

        If `single` is True (default), return only the first article.
        """
        qs = self.filter(relevant_paths__path__iregex=get_path_regex(path))
        if multiple:
            return qs
        return qs.first()

    def top_three_popular(self):
        """Return the top three most viewed articles."""
        return self.order_by("-views").only(*get_only_filters())[:3]

    def top_three_highlighted(self):
        """Return the top three most viewed articles which are highlighted."""
        return self.filter(highlighted=True).order_by("-views").only(*get_only_filters())[:3]

    def top_three_recent(self):
        """Return the top three most recent articles."""
        return self.order_by("-created").only(*get_only_filters())[:3]


class DjangoHelpArticleManager(BASE_MANAGER):
    """Manager for DjangoHelpArticle."""

    def get_queryset(self):
        """Return the queryset for this manager."""
        return super().get_queryset().all()


class DjangoHelpArticle(BASE_MODEL):
    """Stores articles."""

    title = TranslatedField(
        models.CharField(
            _("Title"),
            max_length=30,
            blank=True,
            help_text=_("The title of this article."),
        ),
        EXTRA_LANGUAGES,
    )
    subtitle = TranslatedField(
        models.CharField(
            _("Subtitle"),
            max_length=70,
            blank=True,
            help_text=_("A subtitle for this article."),
        ),
        EXTRA_LANGUAGES,
    )
    slug = models.SlugField(
        _("Slug"),
        max_length=50,
        unique=True,
        help_text=_("A web address friendly version of the title."),
    )
    category = BASE_FOREIGN_KEY_FIELD(
        DjangoHelpCategory,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="articles",
        help_text=_("The category this article belongs to, if any."),
    )

    article_content = TranslatedField(
        MarkdownxField(
            _("Article Content"),
            help_text=_("The content of this article. Markdown is supported."),
            blank=True,
        ),
        EXTRA_LANGUAGES,
    )

    views = models.PositiveIntegerField(
        default=0,
        help_text=_("The number of times this article has been viewed."),
    )
    icon = models.CharField(
        max_length=50,
        blank=True,
        help_text=_("The icon and text color to use for this article."),
        default="fa-circle-info text-success",
    )
    public = models.BooleanField(
        default=True,
        help_text=_("Check this box to make this article public."),
    )
    highlighted = models.BooleanField(
        default=False,
        help_text=_("Check this box to highlight this article on the index page."),
    )

    intended_entity_type = models.CharField(
        max_length=20,
        choices=IntendedEntityType.choices,
        default=IntendedEntityType.ANY,
        help_text=_("The type of entity this article applies to."),
    )

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    objects = DjangoHelpArticleManager.from_queryset(DjangoHelpArticleQuerySet)()

    tags = TaggableManager(
        through=TaggedArticles,
        blank=True,
        help_text=_("A comma-separated list of tags."),
    )

    class Meta(BASE_MODEL.Meta if hasattr(BASE_MODEL, "Meta") else object):
        """Meta options for DjangoHelpArticle."""

        verbose_name = _("DjangoHelp Article")
        verbose_name_plural = _("DjangoHelp Articles")
        ordering = [f"title_{settings.LANGUAGE_CODE}"]  # Default to the primary language
        indexes = [
            models.Index(fields=["slug"]),
        ]

    def clean(self):
        """Ensure the intended entity type matches the category's intended entity type."""
        if self.category and not (
            self.intended_entity_type == self.category.intended_entity_type
            or self.intended_entity_type == IntendedEntityType.ANY
            or self.category.intended_entity_type == IntendedEntityType.ANY
        ):
            validation_error_message = _(
                f"The intended entity type of this article must match the category's intended entity "
                f"type or be any. {self.intended_entity_type=} {self.category.intended_entity_type=}."
            )
            raise ValidationError(validation_error_message)
        super().clean()

    def __str__(self):
        return self.title

    def add_relevant_path(self, path: str):
        """Add the specified path to this article."""
        RelevantPath.objects.create(path=path, article=self)

    def content_preview(self, length=100):
        """Return a preview of the content, truncated to the specified length."""
        return self.article_content[:length] + "..." if len(self.article_content) > length else self.article_content


class RelevantPath(BASE_MODEL):
    """Stores a relevant path for an article.

    A relevant path is a path that is relevant to the article. For example, if the article is about
    creating a new provider, then the relevant path might be `/providers/create`.

    Wildcards are supported. For example, if the article is about creating a new provider, then the relevant
    path might be `/providers/*`.
    """

    path = models.CharField(
        max_length=200,
        help_text=_("The relevant path for this article. Wildcards are supported."),
    )
    article = BASE_FOREIGN_KEY_FIELD(
        DjangoHelpArticle,
        on_delete=models.CASCADE,
        related_name="relevant_paths",
    )

    class Meta(BASE_MODEL.Meta if hasattr(BASE_MODEL, "Meta") else object):
        """Meta options for RelevantPath."""

        verbose_name = _("Relevant Path")
        verbose_name_plural = _("Relevant Paths")
        ordering = ["path"]
        indexes = [
            models.Index(fields=["path"]),
        ]

    def __str__(self):
        return str(self.path)

    @property
    def is_wildcard(self):
        """Return True if the path is a wildcard path."""
        return "*" in list(self.path)


class ArticleUpload(BASE_MODEL):
    """Stores uploaded files for articles."""

    upload = models.FileField(upload_to="uploads/%Y/%m/%d/")

    class Meta:
        """Meta options for ArticleUpload."""

        managed = False  # No database table creation or deletion operations will be performed for this model.
        verbose_name = _("DjangoHelp Article Upload")
        verbose_name_plural = _("DjangoHelp Article Uploads")

    @property
    def filename(self):
        """Return the filename of the upload."""
        return self.upload.name.split("/")[-1]

    @property
    def extension(self):
        """Return the extension of the upload."""
        return self.filename.split(".")[-1]

    def __str__(self):
        return self.filename



"""Utility functions for exporting DjangoHelpArticle instances to Markdown or ZIP."""

import io
import logging
import zipfile

from django.conf import settings
from django.db.models import QuerySet
from django.http import HttpResponse
from translated_fields import to_attribute

from django_help.models import DjangoHelpArticle


logger = logging.getLogger(__name__)


def article_to_markdown(article: DjangoHelpArticle) -> str:
    """Convert a DjangoHelpArticle instance to a Markdown string for each language."""
    tags = ",\n  - ".join(tag.name for tag in article.tags.all())
    relevant_paths = ", ".join(rp.path for rp in article.relevant_paths.all())

    metadata = {
        "id": str(article.id),
        "slug": article.slug,
        "views": article.views,
        "icon": article.icon,
        "public": article.public,
        "highlighted": article.highlighted,
        "intended_entity_type": article.intended_entity_type,
        "created": article.created.strftime("%Y-%m-%d %H:%M:%SZ"),
        "modified": article.modified.strftime("%Y-%m-%d %H:%M:%SZ"),
        "relevant_paths": relevant_paths,
        "tags": tags,
    }

    contents_dict = {}

    for lang_code, _ in settings.LANGUAGES:
        prefix = to_attribute("title", lang_code)
        metadata[f"title_{lang_code}"] = getattr(article, prefix, "")
        metadata[f"subtitle_{lang_code}"] = getattr(article, to_attribute("subtitle", lang_code), "")
        if article.category:
            metadata[f"category_title_{lang_code}"] = getattr(article.category, to_attribute("title", lang_code), "")
            metadata[f"category_subtitle_{lang_code}"] = getattr(
                article.category, to_attribute("subtitle", lang_code), ""
            )

        contents_dict[f"article_content_{lang_code}"] = getattr(article, to_attribute("article_content", lang_code), "")

    headers = "\n".join(f"{key}: {value}" for key, value in metadata.items())
    contents = "\n".join(
        f"# Translation {lang_code}\n{contents_dict[f'article_content_{lang_code}']}"
        for lang_code, _ in settings.LANGUAGES
    )

    return f"---\n{headers}\n---\n\n{contents}"


def export_articles_to_markdown_or_zip(queryset: QuerySet) -> HttpResponse:
    """Export the given queryset of DjangoHelpArticle instances to Markdown (.md) and/or ZIP (if multiple)."""
    if queryset.count() == 1:
        article = queryset.first()
        response = HttpResponse(article_to_markdown(article), content_type="text/markdown")
        response["Content-Disposition"] = f'attachment; filename="{article.slug}.md"'
        return response

    buffer = io.BytesIO()
    with zipfile.ZipFile(buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for article in queryset:
            md_content = article_to_markdown(article)
            zip_file.writestr(f"{article.slug}.md", md_content)
    buffer.seek(0)

    response = HttpResponse(buffer, content_type="application/zip")
    response["Content-Disposition"] = 'attachment; filename="articles.zip"'
    return response



"""Regex utilities for Django."""

import logging
from typing import Union

from django.http import HttpRequest


logger = logging.getLogger(__name__)


def get_path_regex(request_or_path: Union[HttpRequest, str]) -> str:
    """Returns a regex pattern that matches the current path."""

    # If the input is an HttpRequest, get the current path
    if isinstance(request_or_path, HttpRequest):
        current_path = request_or_path.path
    else:
        current_path = request_or_path

    # Convert the path to a regex pattern, replacing '*' with '.*'
    # and anchoring it to match the entire string
    return "^" + current_path.replace("*", ".*") + "$"



"""Utility functions for importing articles from Markdown files or ZIP archives."""

import logging
import os
import tempfile
import zipfile
from typing import Dict
from typing import Optional
from typing import Union

import frontmatter
from django.conf import settings
from django.core.files.uploadhandler import TemporaryFileUploadHandler
from django.db import transaction
from django.http import HttpRequest
from django.shortcuts import redirect
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.csrf import csrf_protect
from translated_fields import to_attribute

from django_help.forms import ArticleUploadForm
from django_help.models import DjangoHelpArticle
from django_help.models import DjangoHelpCategory


logger = logging.getLogger(__name__)


class ArticleManager:
    """Class to manage the creation and updating of DjangoHelpArticle instances from metadata and Markdown content."""

    def __init__(self):
        self.languages = settings.LANGUAGES

    def set_defaults(self, metadata: Dict) -> Dict:
        """Set default values from metadata."""
        defaults = {
            "slug": metadata["slug"],
            "views": metadata.get("views", 0) or 0,
            "public": metadata.get("public", True) if metadata.get("public") is not None else True,
            "highlighted": metadata.get("highlighted", False) if metadata.get("highlighted") is not None else False,
            "intended_entity_type": metadata.get("intended_entity_type", "any") or "any",
        }

        for lang_code, _ in self.languages:
            for field in ["title", "subtitle", "article_content"]:
                key = f"{field}_{lang_code}"
                defaults[to_attribute(field, lang_code)] = metadata.get(key, "") or ""
        return defaults

    def manage_categories(self, metadata: Dict, defaults: Dict) -> None:
        """Manage category creation or fetching.

        Uses the primary language title to create or fetch the category, and then sets the category for the article.
        """
        category_content = {}
        for lang_code, _ in self.languages:
            category_content[f"title_{lang_code}"] = metadata.get(f"category_title_{lang_code}", "") or ""
            category_content[f"subtitle_{lang_code}"] = metadata.get(f"category_subtitle_{lang_code}", "") or ""

        primary_language = settings.LANGUAGE_CODE
        if category_content.get(f"title_{primary_language}") is not None:
            category, _ = DjangoHelpCategory.objects.get_or_create(
                **{to_attribute("title", primary_language): category_content[f"title_{primary_language}"]},
                defaults={
                    key: val
                    for key, val in category_content.items()
                    if val and not key.startswith(f"title_{primary_language}")
                },
            )
            defaults["category"] = category

    def manage_relevant_paths(self, article: DjangoHelpArticle, metadata: Dict) -> None:
        """Clear and update relevant paths."""
        if "relevant_paths" in metadata and metadata["relevant_paths"] is not None:
            article.relevant_paths.all().delete()
            for path in metadata["relevant_paths"]:
                article.relevant_paths.get_or_create(path=path)

    def manage_tags(self, article: DjangoHelpArticle, metadata: Dict) -> None:
        """Update article tags."""
        tags = metadata.get("tags")
        if tags is not None:
            article.tags.set(tags, clear=True)

    def update_or_create_article(self, metadata: Dict) -> tuple[Optional[DjangoHelpArticle], bool]:
        """Update or create a DjangoHelpArticle instance from metadata."""
        if "id" not in metadata or "slug" not in metadata:
            logger.error("Required fields 'id' or 'slug' are missing.")
            return None, False

        with transaction.atomic():
            defaults = self.set_defaults(metadata)

            self.manage_categories(metadata, defaults)

            article, created = DjangoHelpArticle.objects.update_or_create(id=metadata["id"], defaults=defaults)
            self.manage_relevant_paths(article, metadata)
            self.manage_tags(article, metadata)

        return article, created

    def read_markdown_content(self, markdown_file: Union[str, tempfile.SpooledTemporaryFile]) -> str:
        """Read markdown content from a file or file-like object."""
        if isinstance(markdown_file, str):
            with open(markdown_file, "r", encoding="utf-8") as file:
                return file.read()
        else:
            return markdown_file.read().decode("utf-8")

    def extract_language_contents(self, content: str) -> Dict:
        """Extract and map content to languages based on headers in the markdown."""
        contents = content.split("\n# Translation ")
        logger.info("Extracted %s language contents from markdown.", len(contents))
        for idx, content in enumerate(contents):
            logger.info("Content %s: %s...", idx, content[:100])

        language_contents = {}

        # Handle the default language content
        default_lang_code = settings.LANGUAGE_CODE
        if contents:
            language_contents[default_lang_code] = contents[0].strip()

        # Handle translated content
        for content in contents[1:]:
            header_end_idx = content.find("\n")
            if header_end_idx != -1:
                lang_code = content[:header_end_idx].strip()
                content_body = content[header_end_idx + 1 :].strip()  # Start reading content after the newline
                if lang_code in dict(self.languages).keys():  # Ensure the language code is recognized
                    language_contents[lang_code] = content_body

        return language_contents

    def process_markdown_metadata(self, markdown_content: str) -> Optional[Dict]:
        """Parse markdown content and prepare metadata including language-specific content."""
        try:
            post = frontmatter.loads(markdown_content)
            metadata = post.metadata
            language_contents = self.extract_language_contents(post.content)
            for lang_code, content in language_contents.items():
                metadata[f"article_content_{lang_code}"] = content
            return metadata
        except KeyError as e:
            logger.error("Error parsing frontmatter: %s", e)
            return None

    def process_md_file(self, markdown_file: Union[str, tempfile.SpooledTemporaryFile]) -> bool:
        """Process a markdown file and create or update a DjangoHelpArticle instance."""
        markdown_content = self.read_markdown_content(markdown_file)
        logger.info("Processing markdown content starting with: %s...", markdown_content[:100])

        metadata = self.process_markdown_metadata(markdown_content)
        if metadata is None:
            return False

        self.update_or_create_article(metadata)
        return True

    def handle_zip_file(self, zip_file: Union[str, tempfile.SpooledTemporaryFile]) -> None:
        """Extract and process each Markdown file in a ZIP archive."""
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            with tempfile.TemporaryDirectory() as tempdir:
                zip_ref.extractall(tempdir)
                for file_name in zip_ref.namelist():
                    if file_name.lower().endswith(".md"):
                        full_path = os.path.join(tempdir, file_name)
                        self.process_md_file(full_path)
                    else:
                        logger.warning("Skipped non-Markdown file: %s", file_name)


@csrf_exempt
def upload_articles(request: HttpRequest):
    """Insert upload handler before processing uploaded files.

    We must ensure here that uploaded files are handled as temporary files
    See: https://docs.djangoproject.com/en/5.1/topics/http/file-uploads/#modifying-upload-handlers-on-the-fly
    """
    request.upload_handlers.insert(0, TemporaryFileUploadHandler(request))
    return _upload_articles(request)


@csrf_protect
def _upload_articles(request: HttpRequest):
    """Handles the uploading of articles in Markdown or ZIP format by administrators.

    Accessible only via POST method in an admin-specific view.
    """
    if request.method != "POST":
        return _handle_get_request(request)

    return _handle_post_request(request)


def _handle_get_request(request: HttpRequest):
    """Render the empty form for GET requests."""
    template = "django_help/admin/upload_articles.html"
    form = ArticleUploadForm()
    return render(request, template, {"form": form})


def _handle_post_request(request: HttpRequest):
    """Process the file upload through POST request."""
    template = "django_help/admin/upload_articles.html"
    form = ArticleUploadForm(request.POST, request.FILES)
    if not form.is_valid():
        return render(request, template, {"form": form})

    uploaded_file = form.cleaned_data.get("upload")
    if not uploaded_file:
        # Handle the case where the file is not provided
        form.add_error("upload", "Please upload a file.")
        return render(request, template, {"form": form})

    # Process the file based on its type
    article_manager = ArticleManager()
    if uploaded_file.name.endswith(".zip"):
        article_manager.handle_zip_file(uploaded_file)
    else:
        article_manager.process_md_file(uploaded_file)

    return redirect("admin:index")



"""This module contains the choices for the models in the django_help app."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_help.app_settings import INTENDED_ENTITY_TYPE


class IntendedEntityType(models.TextChoices):
    """Stores the type of entity that a category or article is intended for."""

    ANY = "any", _("Any")
    for key, (value, display) in INTENDED_ENTITY_TYPE.items():
        locals()[key] = (value, display)



"""Admin configuration for django_help."""

import logging

from django.conf import settings
from django.contrib import admin
from django.urls import path
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _
from markdownx.admin import MarkdownxModelAdmin
from translated_fields import TranslatedFieldAdmin

from django_help.models import DjangoHelpArticle
from django_help.models import DjangoHelpCategory
from django_help.models import RelevantPath
from django_help.models import TaggedArticles
from django_help.utils.exporting import export_articles_to_markdown_or_zip
from django_help.utils.importing import upload_articles


logger = logging.getLogger(__name__)

installed_languages = [lang_code for lang_code, _ in settings.LANGUAGES]


@admin.register(TaggedArticles)
class TaggedArticlesAdmin(admin.ModelAdmin):
    """Admin configuration for TaggedArticles instances."""

    list_display = ("tag",)
    search_fields = ("tag",)


class RelevantPathInline(admin.TabularInline):
    """Inline for RelevantPath instances."""

    model = RelevantPath
    extra = 0


@admin.action(description="Export selected articles to Markdown (.md) and/or ZIP")
def export_articles(modeladmin, request, queryset):
    """Action to export the selected articles to Markdown (.md) and/or ZIP."""
    return export_articles_to_markdown_or_zip(queryset)


@admin.register(DjangoHelpCategory)
class DjangoHelpCategoryAdmin(TranslatedFieldAdmin, admin.ModelAdmin):
    """Admin configuration for DjangoHelpCategory instances."""

    list_display = (
        f"title_{get_language()}",
        "slug",
        f"subtitle_{get_language()}",
        "icon",
        "intended_entity_type",
        "public",
        "created",
        "modified",
    )
    list_filter = ("public", "intended_entity_type")
    search_fields = (
        *[f"title_{lang_code}" for lang_code in installed_languages],
        *[f"subtitle_{lang_code}" for lang_code in installed_languages],
        *[f"description_{lang_code}" for lang_code in installed_languages],
        "slug",
    )
    readonly_fields = ("created", "modified")
    fieldsets = (
        (_("title"), {"fields": DjangoHelpCategory.title.fields}),
        (_("subtitle"), {"fields": DjangoHelpCategory.subtitle.fields}),
        (_("description"), {"fields": DjangoHelpCategory.description.fields}),
        (_(""), {"fields": ["slug", "icon", "intended_entity_type", "public", "created", "modified"]}),
    )
    prepopulated_fields = {"slug": (f"title_{get_language()}",)}


@admin.register(DjangoHelpArticle)
class DjangoHelpArticleAdmin(TranslatedFieldAdmin, MarkdownxModelAdmin):
    """Admin configuration for DjangoHelpArticle instances."""

    list_display = (
        f"title_{get_language()}",
        "slug",
        f"subtitle_{get_language()}",
        "category",
        "icon",
        "public",
        "highlighted",
        "intended_entity_type",
        "views",
        "tag_list",
        "created",
        "modified",
    )
    list_filter = ("public", "category", "intended_entity_type", "highlighted")
    search_fields = (
        *[f"title_{lang_code}" for lang_code in installed_languages],
        *[f"subtitle_{lang_code}" for lang_code in installed_languages],
        *[f"article_content_{lang_code}" for lang_code in installed_languages],
        "slug",
        "tags",
    )
    readonly_fields = ("created", "modified")
    inlines = [RelevantPathInline]
    actions = [export_articles]
    fieldsets = (
        (_("title"), {"fields": DjangoHelpArticle.title.fields}),
        (_("subtitle"), {"fields": DjangoHelpArticle.subtitle.fields}),
        (_("article_content"), {"fields": DjangoHelpArticle.article_content.fields}),
        (_(""), {"fields": ["category", "slug", "icon", "public", "highlighted", "intended_entity_type", "tags"]}),
    )
    prepopulated_fields = {"slug": (f"title_{get_language()}",)}

    def changelist_view(self, request, extra_context=None):
        """Add the upload button to the changelist view."""
        extra_context = extra_context or {}
        extra_context["show_upload_button"] = True
        extra_context["upload_url"] = "upload/"
        return super().changelist_view(request, extra_context=extra_context)

    def get_urls(self):
        """Add the upload view to the admin."""
        urls = super().get_urls()
        my_urls = [
            path("upload/", self.admin_site.admin_view(upload_articles), name="upload_articles"),
        ]
        return my_urls + urls

    def get_queryset(self, request):
        """Prefetch tags for the queryset."""
        return super().get_queryset(request).prefetch_related("tags")

    def tag_list(self, obj):
        """Return a comma-separated list of tags for the object."""
        return ", ".join(o.name for o in obj.tags.all())



"""Settings for the django_help app.

Use the DJANGO_HELP_CONFIG setting to configure the app. For example:

.. code-block:: python

    DJANGO_HELP_CONFIG = {
        "BASE_MANAGER": "myapp.models.MyManager",
        "BASE_QUERYSET": "myapp.models.MyQuerySet",
        "BASE_MODEL": "myapp.models.MyModel",
        "BASE_FOREIGN_KEY": "myapp.models.MyForeignKey",
        "EXTRA_LANGUAGES": {
            "es": {"blank": False},  # Spanish fields must be filled out.
            "fr": {"blank": True},  # French fields can be left blank.
        },
        "AUTHORIZE_USER_TO_VIEW_ARTICLE": "myapp.utils.authorize_user_to_view_article",
        "INTENDED_ENTITY_TYPE": {
            "PERSON": ("person", "Person"),
            "ORGANIZATION": ("organization", "Organization"),
        },
    }
"""

from django.conf import settings
from django.db import models

from django_help.utils.validation import user_authorized_to_view


django_help_config = getattr(settings, "DJANGO_HELP_CONFIG", {})

# Specifies the base form for all forms in the app.
BASE_MANAGER = django_help_config.get("BASE_MANAGER", models.Manager)

# Specifies the base queryset for all models in the app.
BASE_QUERYSET = django_help_config.get("BASE_QUERYSET", models.QuerySet)

# Specifies the base model for all models in the app.
BASE_MODEL = django_help_config.get("BASE_MODEL", models.Model)

# Specifies the model field for foreign keys in the app.
BASE_FOREIGN_KEY_FIELD = django_help_config.get("BASE_FOREIGN_KEY", models.ForeignKey)

# Specifies extra languages to use in the app.
EXTRA_LANGUAGES = django_help_config.get("EXTRA_LANGUAGES", {})

# A function for authorizing users to view articles.
# The function should take a request and an article as arguments and return True if the user is authorized to view the
# article, and False otherwise.
AUTHORIZE_USER_TO_VIEW_ARTICLE = django_help_config.get("AUTHORIZE_USER_TO_VIEW_ARTICLE", user_authorized_to_view)

# The default intended entity type for articles and categories.
# This is a tuple of the form (code, name).
# ("any", "Any") is the default value, which can be used as `IntendedEntityType.ANY`.
# This is optional, and can be used for filtering articles and categories by intended entity type.
# How you define the intended entity type is up to you.
INTENDED_ENTITY_TYPE = django_help_config.get("INTENDED_ENTITY_TYPE", {})



"""Views for django_help."""

import logging

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator
from django.db.models import Count
from django.db.models import Q
from django.http import Http404
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from translated_fields import to_attribute

from django_help.app_settings import AUTHORIZE_USER_TO_VIEW_ARTICLE
from django_help.forms import DjangoHelpArticleForm
from django_help.models import DjangoHelpArticle
from django_help.models import DjangoHelpCategory
from django_help.utils.queryset import annotate_with_values_fields
from django_help.utils.queryset import filter_articles
from django_help.utils.queryset import filter_user_can_view


logger = logging.getLogger(__name__)


@login_required
def index(request: HttpRequest) -> TemplateResponse:
    """The main django_help page.

    Provides:
    - a search box
    - an option for admins to create new articles
    - a list of popular articles
    - a list of highlighted articles, and
    - a list of recent articles.
    """
    template = "django_help/pages-index.html"
    context = {}
    context["url"] = request.path
    queryset = DjangoHelpArticle.objects.all()

    base_article_qs = filter_user_can_view(request, queryset)

    categories = DjangoHelpCategory.objects.annotate(
        number_of_articles=Count("articles", filter=Q(articles__public=True))
    )
    categories = filter_user_can_view(request, categories)

    context["popular_articles"] = base_article_qs.top_three_popular()
    context["highlighted_articles"] = base_article_qs.top_three_highlighted()
    context["recent_articles"] = base_article_qs.top_three_recent()
    context["categories"] = categories.order_by(to_attribute("title"))

    if not request.user.is_staff:
        queryset = queryset.public()

    queryset = queryset.order_by("-modified")

    search_query = request.GET.get("search", "").strip()
    if search_query:
        search_keywords = search_query.lower().split(",")
        query = list(map(str.strip, search_keywords))
        queryset = queryset.search(query)

        # Paginate the queryset
        paginator = Paginator(queryset, 9)
        page_number = request.GET.get("page", 1)
        page_obj = paginator.get_page(page_number)
        context["articles"] = page_obj
        context["is_paginated"] = page_obj.has_other_pages()

    return TemplateResponse(request, template, context)


@login_required
def category(request: HttpRequest, category_slug: str) -> TemplateResponse:
    """The category page for django_help.

    Provides:
    - a list of categories
    - a list of articles in the current category
    """
    template = "django_help/pages-category.html"
    context = {}
    current_category = get_object_or_404(DjangoHelpCategory, slug=category_slug)

    if request.user.is_staff:
        articles_in_current_category = current_category.articles.all()
        categories = DjangoHelpCategory.objects.all()
    else:
        articles_in_current_category = filter_user_can_view(request, current_category.articles.all())
        categories = filter_user_can_view(request, DjangoHelpCategory.objects.all())

    # Annotate with `title` field
    values_fields = ["title"]
    articles_in_current_category = annotate_with_values_fields(articles_in_current_category, values_fields)
    categories = annotate_with_values_fields(categories, values_fields)

    articles_in_current_category = articles_in_current_category.order_by("title").values("title", "slug")
    categories = categories.order_by("title").values("title", "slug")

    context["current_category"] = current_category
    context["articles_in_current_category"] = articles_in_current_category
    context["categories"] = categories

    return TemplateResponse(request, template, context)


@login_required
def view_article(request: HttpRequest, article_slug: str) -> TemplateResponse:
    """Uses htmx to load a specific article instance based on the slug."""
    if "HTTP_HX_REQUEST" in request.META and request.META["HTTP_HX_REQUEST"] == "true":
        context = {}
        template = "django_help/fragments/modal_content.html"

        article = DjangoHelpArticle.objects.filter(slug=article_slug).first()
        AUTHORIZE_USER_TO_VIEW_ARTICLE(request, article)

        if request.user.is_staff and article is None:
            template = "django_help/fragments/create_article.html"
            form = DjangoHelpArticleForm()
            context["form"] = form
        else:
            context["article"] = article
            if article is None:
                logger.error("Article with slug %s not found", article_slug)
                raise Http404

        return TemplateResponse(request, template, context)
    logger.error("Request to view article %s was not an htmx request", article_slug)
    raise Http404


@login_required
def get_articles(request: HttpRequest, *, category: str = None, slug: str = None, tag: str = None) -> TemplateResponse:
    """Fetches articles based on category, slug, or tag."""
    logger.info("Called get_articles for: %s, %s, %s, %s", category, slug, tag, request.path)
    context = {}
    articles = filter_articles(request, category, slug, tag)

    if articles.exists():
        articles = articles.distinct()
    else:
        logger.warning("No articles found for: %s, %s, %s, %s", category, slug, tag, request.path)
        articles = filter_user_can_view(request, DjangoHelpArticle.objects.all().distinct())

    context["article"] = first_article = articles.first()
    context["articles"] = articles.exclude(id=first_article.id).top_three_popular()

    template = "django_help/fragments/modal_content.html"

    return TemplateResponse(request, template, context)


@login_required
@staff_member_required
def create_article(request):
    """Creates a new article instance."""
    template = "django_help/fragments/create_article.html"
    context = {}

    if request.method == "POST":
        form = DjangoHelpArticleForm(request.POST)

        if form.is_valid():
            article = form.save()
            context["article"] = article
            template = "django_help/fragments/modal_content.html"
            messages.success(request, "Article created successfully!")
            return TemplateResponse(request, template, context)
        else:
            context["form"] = form
            return TemplateResponse(request, template, context)

    form = DjangoHelpArticleForm()

    context["form"] = form
    return TemplateResponse(request, template, context)


@login_required
@staff_member_required
def edit_article(request, article_slug):
    """Edits an existing article instance."""
    article = get_object_or_404(DjangoHelpArticle, slug=article_slug)

    context = {}
    context["article"] = article

    template = "django_help/fragments/edit_article.html"
    success_template = "django_help/fragments/modal_content.html"

    if request.method == "POST":
        form = DjangoHelpArticleForm(request.POST, instance=article)
        if form.is_valid():
            form.save()
            messages.success(request, "Article modified successfully!")
            return TemplateResponse(request, success_template, context)
        else:
            context["form"] = form
            return TemplateResponse(request, template, context)
    else:
        form = DjangoHelpArticleForm(instance=article)
        context["form"] = form
        return TemplateResponse(request, template, context)



"""Forms for django_help."""

from crispy_forms.bootstrap import PrependedText
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Div
from crispy_forms.layout import Field
from crispy_forms.layout import Layout
from django import forms
from django.conf import settings
from markdownx.fields import MarkdownxFormField
from markdownx.widgets import MarkdownxWidget
from translated_fields.utils import language_code_formfield_callback

from django_help.app_settings import EXTRA_LANGUAGES
from django_help.models import ArticleUpload
from django_help.models import DjangoHelpArticle
from django_help.models import RelevantPath


class RelevantPathForm(forms.ModelForm):
    """Form for creating RelevantPath instances."""

    class Meta:
        """Meta class for RelevantPathForm."""

        model = RelevantPath
        fields = ("path",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()

        # Use FormHelper to add form tags and set form attributes
        self.helper.form_tag = False
        self.helper.disable_csrf = True
        self.helper.form_class = "form-horizontal"
        self.helper.label_class = "col-lg-2"
        self.helper.field_class = "col-lg-8"


class DjangoHelpArticleForm(forms.ModelForm):
    """Form for creating and editing articles."""

    formfield_callback = language_code_formfield_callback

    class Meta:
        """Meta class for DjangoHelpArticleForm."""

        model = DjangoHelpArticle
        fields = (
            *DjangoHelpArticle.title.fields,
            *DjangoHelpArticle.subtitle.fields,
            "category",
            *DjangoHelpArticle.article_content.fields,
            "icon",
            "public",
            "highlighted",
            "intended_entity_type",
            "tags",
        )
        widgets = {
            f"article_content_{lang_code}": MarkdownxWidget(
                attrs={
                    "id": f"article_content_{lang_code}",
                    "class": "form-control",
                    "rows": 5,
                }
            )
            for lang_code, _ in settings.LANGUAGES
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Create `title_{lang_code}`, `subtitle_{lang_code}`, and `article_content_{lang_code}` for each language.
        for lang_code, lang_name in settings.LANGUAGES:
            # For each below, if blank is not specified for the language in EXTRA_LANGUAGES, default to required=True.
            self.fields[f"title_{lang_code}"] = forms.CharField(
                label=f"Title ({lang_name})",
                required=EXTRA_LANGUAGES.get(lang_code, {}).get("blank", True),
            )
            self.fields[f"subtitle_{lang_code}"] = forms.CharField(
                label=f"Subtitle ({lang_name})",
                required=EXTRA_LANGUAGES.get(lang_code, {}).get("blank", True),
            )
            self.fields[f"article_content_{lang_code}"] = MarkdownxFormField(
                label=f"Article Content ({lang_name})",
                required=EXTRA_LANGUAGES.get(lang_code, {}).get("blank", True),
            )

        # tags should not be required
        self.fields["tags"].required = False

        self.helper = FormHelper()

        # Use FormHelper to add form tags and set form attributes
        self.helper.form_tag = False
        self.helper.disable_csrf = True
        self.helper.form_class = "form-horizontal"
        self.helper.label_class = "col-lg-2"
        self.helper.field_class = "col-lg-10"

        self.helper.layout = Layout(
            # Add a Div for the title, subtitle, and article_content fields for each language
            *[
                Div(
                    Field(f"title_{lang_code}", css_class="form-control"),
                    css_class="col-lg-12 mb-3",
                )
                for lang_code, lang_name in settings.LANGUAGES
            ],
            *[
                Div(
                    Field(f"subtitle_{lang_code}", css_class="form-control"),
                    css_class="col-lg-12 mb-3",
                )
                for lang_code, lang_name in settings.LANGUAGES
            ],
            Div(
                Field("category", css_class="form-control"),
                css_class="col-lg-12 mb-3",
            ),
            *[
                Div(
                    Field(f"article_content_{lang_code}", css_class="form-control"),
                    css_class="col-lg-12 mb-3",
                )
                for lang_code, lang_name in settings.LANGUAGES
            ],
            Div(
                Field("icon", css_class="form-control"),
                css_class="col-lg-12 mb-3",
            ),
            Div(
                Field("public", css_class="form-check-input"),
                css_class="col-lg-12 mb-3",
            ),
            Div(
                Field("highlighted", css_class="form-check-input"),
                css_class="col-lg-12 mb-3",
            ),
            Div(
                Field("intended_entity_type", css_class="form-control"),
                css_class="col-lg-12 mb-3",
            ),
            Div(
                PrependedText("tags", "#", placeholder="tags", css_class="form-control"),
                css_class="col-lg-12 mb-4",
            ),
        )


class ArticleUploadForm(forms.ModelForm):
    """Form for uploading articles in admin."""

    class Meta:
        """Meta class for ArticleUploadForm."""

        model = ArticleUpload
        fields = ["upload"]



"""App configuration for django-help."""

from django.apps import AppConfig


class DjangoHelpConfig(AppConfig):
    """App configuration for django-help."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_help"

    def ready(self):
        """Run when the app is ready."""
        from django_help import checks  # pylint: disable=C0415 disable=W0611



"""Queryset utilities for the Django Help app."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Union
from urllib.parse import unquote

from django.db.models import F
from django.http import HttpRequest
from translated_fields import to_attribute

from django_help.models import DjangoHelpArticle


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from django_help.models import DjangoHelpArticleQuerySet
    from django_help.models import DjangoHelpCategoryQuerySet


def filter_user_can_view(
    request: HttpRequest, articles_or_categories: Union[DjangoHelpArticleQuerySet, DjangoHelpCategoryQuerySet]
) -> Union[DjangoHelpArticleQuerySet, DjangoHelpCategoryQuerySet]:
    """Filters QuerySet to return only public articles or categories based on the user's entity context."""
    if request.user.is_staff:
        return articles_or_categories
    if request.entity_context.id == request.provider_context.id:
        return articles_or_categories.intended_for_providers_or_any().public()
    return articles_or_categories.intended_for_consumers_or_any().public()


def filter_articles(request: HttpRequest, category: str, slug: str, tag: str) -> DjangoHelpArticleQuerySet:
    """Filters articles based on user permissions and provided parameters.

    Note: This is a helper function for get_articles.
    """
    logger.info(
        "Called get_filtered_articles for: %s, %s, %s, %s, %s, %s",
        category,
        type(category),
        slug,
        type(slug),
        tag,
        request.path,
    )

    articles = filter_user_can_view(request, DjangoHelpArticle.objects.all())

    if tag is not None:
        logger.info("Getting articles for: %s", tag)
        articles = articles.get_for_tag(unquote(tag), multiple=True)
    elif slug is not None:
        logger.info("Getting articles for: %s", slug)
        articles = articles.get_for_slug(unquote(slug), multiple=True)
    elif category is not None:
        logger.info("Getting articles for: %s", category)
        articles = articles.get_for_category(unquote(category), multiple=True)
    else:
        logger.info("Getting articles for: %s", request.path)
        articles = articles.get_for_path(request.path, multiple=True)

    if not request.user.is_staff:
        articles = articles.public()

    return articles


def annotate_with_values_fields(queryset, values_fields: list[str]):
    """Annotates the queryset with the values fields.

    For instance, if values_fields = ["title"], the queryset will be annotated with `title=to_attribute("title")`,
      ensuring the current language is used. e.g.: If the currentl language is 'en', the annotation will convert
      `title_en` to `title`.
    """
    return queryset.annotate(**{field: F(to_attribute(field)) for field in values_fields})



"""URLs for django_help."""

from django.urls import path

from django_help import views


app_name = "django_help"

urlpatterns = [
    path("", views.index, name="index"),
    path("create/", views.create_article, name="create_article"),
    path("category/<slug:category_slug>/", views.category, name="category"),
    path("article/<slug:article_slug>/edit/", views.edit_article, name="edit_article"),
    path("article/<slug:article_slug>/", views.view_article, name="view_article"),
    # The following paths all share the same name "get_articles" and are distinguished by the presence of the kwargs.
    # See: https://docs.djangoproject.com/en/dev/topics/http/urls/#naming-url-patterns
    path(
        "articles/category/<str:category>/",
        views.get_articles,
        name="get_articles",
    ),
    path(
        "articles/slug/<slug:slug>/",
        views.get_articles,
        name="get_articles",
    ),
    path(
        "articles/tag/<str:tag>/",
        views.get_articles,
        name="get_articles",
    ),
    path(
        "articles/",
        views.get_articles,
        name="get_articles",
    ),
]



"""Defines custom template tags for django_help."""

from collections.abc import Iterable

import markdown as md
from django import template
from django.templatetags.static import static
from django.utils.safestring import mark_safe


register = template.Library()

try:
    # For Django 5.1 and above, we can import the querystring tag
    from django.template.defaulttags import querystring  # pylint: disable=W0611
except ImportError:
    # For Django 5.0 and below, we need to define the querystring tag

    @register.simple_tag(name="querystring", takes_context=True)
    def querystring(context, query_dict=None, **kwargs):
        """
        Add, remove, and change parameters of a ``QueryDict`` and return the result
        as a query string. If the ``query_dict`` argument is not provided, default
        to ``request.GET``.

        For example::

            {% querystring foo=3 %}

        To remove a key::

            {% querystring foo=None %}

        To use with pagination::

            {% querystring page=page_obj.next_page_number %}

        A custom ``QueryDict`` can also be used::

            {% querystring my_query_dict foo=3 %}
        """
        if query_dict is None:
            query_dict = context.request.GET
        query_dict = query_dict.copy()
        for key, value in kwargs.items():
            if value is None:
                if key in query_dict:
                    del query_dict[key]
            elif isinstance(value, Iterable) and not isinstance(value, str):
                query_dict.setlist(key, value)
            else:
                query_dict[key] = value
        if not query_dict:
            return ""
        query_string = query_dict.urlencode()
        return f"?{query_string}"


@register.simple_tag
def load_dependencies():
    """Loads the dependencies for django_help."""
    css_url = static("django_help/css/styles.css")
    # return mark_safe(
    #     f"""
    #     <!-- Include AlpineJS -->
    #     <script src="https://cdn.jsdelivr.net/npm/alpinejs@2" defer></script>
    #     <!-- Include htmx -->
    #     <script src="https://unpkg.com/htmx.org/dist/htmx.js"></script>
    #     <!-- Include custom css -->
    #     <link href="{css_url}" rel="stylesheet">
    #     """
    # )
    return mark_safe(
        f"""
        <!-- Include custom css -->
        <link href="{css_url}" rel="stylesheet">
        """
    )


@register.filter(name="markdown")
def markdown_format(text):
    """Formats the given text as markdown and processes internal links."""
    import re

    from django.urls import reverse
    from django.utils.safestring import mark_safe

    def replace_internal_links(match):
        slug = match.group(1)
        title = match.group(2) if match.group(2) else slug
        url = reverse("django_help:view_article", kwargs={"article_slug": slug})
        return f'<a href="{url}" class="internal-link" data-slug="{slug}">{title}</a>'

    # Convert markdown to HTML
    html = md.markdown(text, extensions=["markdown.extensions.fenced_code"])

    # Replace internal links with custom format
    pattern = r"\[\[([^\]|]+)(?:\|([^\]]+))?\]\]"
    html = re.sub(pattern, replace_internal_links, html)

    return mark_safe(html)


@register.inclusion_tag("django_help/fragments/button_slug.html")
def django_help_slug(slug, title=None):
    """Renders a django_help button with the given slug."""
    return {"slug": slug, "title": title}


@register.inclusion_tag("django_help/fragments/button_category.html")
def django_help_category(category, title=None):
    """Renders a django_help button with the given category."""
    return {"category": category, "title": title}


@register.inclusion_tag("django_help/fragments/button_tag.html")
def django_help_tag(tag, title=None):
    """Renders a django_help button with the given tag."""
    return {"tag": tag, "title": title}


@register.inclusion_tag("django_help/fragments/button_current_path.html")
def django_help_current_path(title=None):
    """Renders a django_help button with the given current path."""
    return {"title": title}



# Generated by Django 4.2.16 on 2024-10-05 00:21

from django.db import migrations, models
import django.db.models.deletion
import markdownx.models
import taggit.managers


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ArticleUpload",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("upload", models.FileField(upload_to="uploads/%Y/%m/%d/")),
            ],
            options={
                "verbose_name": "DjangoHelp Article Upload",
                "verbose_name_plural": "DjangoHelp Article Uploads",
                "managed": False,
            },
        ),
        migrations.CreateModel(
            name="DjangoHelpArticle",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "title_en",
                    models.CharField(
                        blank=True, help_text="The title of this article.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "title_es",
                    models.CharField(
                        blank=True, help_text="The title of this article.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "title_fr",
                    models.CharField(
                        blank=True, help_text="The title of this article.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "subtitle_en",
                    models.CharField(
                        blank=True, help_text="A subtitle for this article.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "subtitle_es",
                    models.CharField(
                        blank=True, help_text="A subtitle for this article.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "subtitle_fr",
                    models.CharField(
                        blank=True, help_text="A subtitle for this article.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="A web address friendly version of the title.", unique=True, verbose_name="Slug"
                    ),
                ),
                (
                    "article_content_en",
                    markdownx.models.MarkdownxField(
                        blank=True,
                        help_text="The content of this article. Markdown is supported.",
                        verbose_name="Article Content",
                    ),
                ),
                (
                    "article_content_es",
                    markdownx.models.MarkdownxField(
                        blank=True,
                        help_text="The content of this article. Markdown is supported.",
                        verbose_name="Article Content",
                    ),
                ),
                (
                    "article_content_fr",
                    markdownx.models.MarkdownxField(
                        blank=True,
                        help_text="The content of this article. Markdown is supported.",
                        verbose_name="Article Content",
                    ),
                ),
                (
                    "views",
                    models.PositiveIntegerField(
                        default=0, help_text="The number of times this article has been viewed."
                    ),
                ),
                (
                    "icon",
                    models.CharField(
                        blank=True,
                        default="fa-circle-info text-success",
                        help_text="The icon and text color to use for this article.",
                        max_length=50,
                    ),
                ),
                ("public", models.BooleanField(default=True, help_text="Check this box to make this article public.")),
                (
                    "highlighted",
                    models.BooleanField(
                        default=False, help_text="Check this box to highlight this article on the index page."
                    ),
                ),
                (
                    "intended_entity_type",
                    models.CharField(
                        choices=[("any", "Any")],
                        default="any",
                        help_text="The type of entity this article applies to.",
                        max_length=20,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("modified", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "DjangoHelp Article",
                "verbose_name_plural": "DjangoHelp Articles",
                "ordering": ["title_en"],
            },
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100, unique=True, verbose_name="name")),
                ("slug", models.SlugField(allow_unicode=True, max_length=100, unique=True, verbose_name="slug")),
            ],
            options={
                "verbose_name": "tag",
                "verbose_name_plural": "tags",
            },
        ),
        migrations.CreateModel(
            name="TaggedArticles",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "content_object",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="django_help.djangohelparticle"),
                ),
                (
                    "tag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(app_label)s_%(class)s_items",
                        to="django_help.tag",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tagged Article",
                "verbose_name_plural": "Tagged Articles",
            },
        ),
        migrations.CreateModel(
            name="DjangoHelpCategory",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "title_en",
                    models.CharField(
                        blank=True, help_text="The title of this category.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "title_es",
                    models.CharField(
                        blank=True, help_text="The title of this category.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "title_fr",
                    models.CharField(
                        blank=True, help_text="The title of this category.", max_length=30, verbose_name="Title"
                    ),
                ),
                (
                    "subtitle_en",
                    models.CharField(
                        blank=True, help_text="A subtitle for this category.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "subtitle_es",
                    models.CharField(
                        blank=True, help_text="A subtitle for this category.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "subtitle_fr",
                    models.CharField(
                        blank=True, help_text="A subtitle for this category.", max_length=70, verbose_name="Subtitle"
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="A web address friendly version of the title.", unique=True, verbose_name="Slug"
                    ),
                ),
                (
                    "description_en",
                    models.CharField(
                        blank=True,
                        help_text="A description of this category.",
                        max_length=200,
                        verbose_name="Description",
                    ),
                ),
                (
                    "description_es",
                    models.CharField(
                        blank=True,
                        help_text="A description of this category.",
                        max_length=200,
                        verbose_name="Description",
                    ),
                ),
                (
                    "description_fr",
                    models.CharField(
                        blank=True,
                        help_text="A description of this category.",
                        max_length=200,
                        verbose_name="Description",
                    ),
                ),
                (
                    "icon",
                    models.CharField(
                        blank=True,
                        default="fa-circle-info text-success",
                        help_text="The icon and text color to use for this category.",
                        max_length=50,
                    ),
                ),
                ("public", models.BooleanField(default=True, help_text="Check this box to make this category public.")),
                (
                    "intended_entity_type",
                    models.CharField(
                        choices=[("any", "Any")],
                        default="any",
                        help_text="The type of entity this category applies to.",
                        max_length=20,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("modified", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "DjangoHelp Category",
                "verbose_name_plural": "DjangoHelp Categories",
                "ordering": ["title_en"],
                "indexes": [models.Index(fields=["slug"], name="django_help_slug_57ad1a_idx")],
            },
        ),
        migrations.AddField(
            model_name="djangohelparticle",
            name="category",
            field=models.ForeignKey(
                blank=True,
                help_text="The category this article belongs to, if any.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="articles",
                to="django_help.djangohelpcategory",
            ),
        ),
        migrations.AddField(
            model_name="djangohelparticle",
            name="tags",
            field=taggit.managers.TaggableManager(
                blank=True,
                help_text="A comma-separated list of tags.",
                through="django_help.TaggedArticles",
                to="django_help.Tag",
                verbose_name="Tags",
            ),
        ),
        migrations.CreateModel(
            name="RelevantPath",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "path",
                    models.CharField(
                        help_text="The relevant path for this article. Wildcards are supported.", max_length=200
                    ),
                ),
                (
                    "article",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="relevant_paths",
                        to="django_help.djangohelparticle",
                    ),
                ),
            ],
            options={
                "verbose_name": "Relevant Path",
                "verbose_name_plural": "Relevant Paths",
                "ordering": ["path"],
                "indexes": [models.Index(fields=["path"], name="django_help_path_e58f80_idx")],
            },
        ),
        migrations.AddIndex(
            model_name="djangohelparticle",
            index=models.Index(fields=["slug"], name="django_help_slug_97cf18_idx"),
        ),
    ]
